#!/usr/local/bin/python3

# S.D.G.

# Name: oglycanstring.py
# Description: oglycanstring is a simulator of the enzymes of mucin-type O-glycosylation
# Author: Andrew G. McDonald
# Date: 22 January 2016

# Usage:
# python3 oglycanstring.py <iterations> [options]

# Options:
# 	-a	Use refined enzyme rules that match additional substrate-patterns
# 	-ln	Limit the number of GlcNAc residues to n per glycan
# 	-plist	Choose enzymes to knock out, as a comma-separated list of assignments of the form name=val, 
# 		where name is the short enzyme name and val is 0 (knockout) or 1 (knock-in). All but SAT-2 are enabled 
# 		by default. Example: -pC2Gn-T=0,ST3Gal-I=0
# 	-sstring
#		Set the initial substrate to value of string (default is "T"). Example: -s"[L3]VT"

import sys, re

# define the enzymes used by the simulator as a tuple

enzymes = (
"beta4Gal-T4",
"ppGalNAc-Ts",
"alpha4Fuc-T",
"alpha2Fuc-Ts",
"C2Gn-T",
"C1Gal-T1",
"beta3Gn-T3",
"beta3Gn-T6",
"C2/4Gn-T",
"beta3Gn-T2/3/4/5/7",
"alpha3Fuc-T",
"beta3Gal-T5",
"ST3Gal-I",
"ST6GalNAc-I",
"ST3Gal-III/IV",
"ST6Gal-I",
"ST6GalNAc-III/IV",
"ST6GlcNAc-I",
"Gcnt2",
"CHST4/6",
"GALST2/3",
"GALST4",
"SAT-2",
"alpha3Gal-T",
"alpha3GalNAc-T",
"beta4GalNAc-T",
)


# set default parameter values

p = {}
for e in enzymes:
	p[e] = 1.0

# SAT-2 disabled
p["SAT-2"] = 0.0

limit = 0
# "advanced" mode activates further enzyme-substrate recognition patterns:
#  use command-line switch '-a' to enable
advanced = ""
halt = 0
# halt becomes true (1) if limit > 0 and c > limit
glycan = "T"

# parse the command line

if len(sys.argv)>1:
	maxiter = int(sys.argv[1])
	arglist = sys.argv[2:]
	for item in arglist:
		if item == '-a':
			advanced = '-a'
			print ("Using advanced model")
		if item.startswith('-l'):
			limit = int(item.replace('-l','',1))
			print ("Limiting to level", limit)
		if item.startswith('-p'):
			kolist = item.replace('-p','',1).split(",")
			for ep in kolist:
				pair = ep.split("=")
				p[pair[0]] = int(pair[1])
		if item.startswith('-s'):
			glycan = item.replace('-s','',1)
	
else:
	print ("No args supplied")
	maxiter = 1



# define the remaining hashes (dictionaries)

newp = {}
news = newp
pool = {}

newg = 0
newp[glycan] = glycan

if glycan not in pool:
	pool[glycan] = glycan
	newg += 1
else:
	newg = 0
n = {}
n[glycan] = len(pool)
product = ""
iteration = 0

# function definitions

def react( string, action ):
	global advanced
	if action == "C1Gal-T1":
		string = re.sub('^V(?P<MATCH1>[ST])', '[L3]V\g<MATCH1>', string, 1)
	elif action == "ST6GalNAc-I":
		# recognises Core 3 in addition to Core 1
		string = re.sub('^(?P<MATCH1>(\[[LY]b?3\])?)V(?P<MATCH2>[ST])', '[S6]\g<MATCH1>V\g<MATCH2>', string, 1)
	elif action == "ST3Gal-I":
		string = re.sub('\[L3\]V(?P<MATCH1>[ST])$', '[S3L3]V\g<MATCH1>', string, 1)
	elif action == "ST6GalNAc-III/IV":
		string = re.sub('^\[S3L3\]V(?P<MATCH1>[ST])', '[S6][S3L3]V\g<MATCH1>', string, 1)
	elif action == "ppGalNAc-Ts":
		string = re.sub('^(?P<MATCH1>[ST])$', 'V\g<MATCH1>', string, 1)
	elif action == "beta3Gn-T3":
		string = re.sub('\[L3\]V(?P<MATCH1>[ST])$', '[Y3L3]V\g<MATCH1>', string, 1)
	elif action == "ST6GlcNAc-I":
		string = re.sub('^\[Y3L3\]V(?P<MATCH1>[ST])$', '[S6Y3L3]V\g<MATCH1>', string, 1)
	elif action == "beta3Gal-T5":
		string = re.sub('\[(?P<MATCH1>(\[s6\])?)Y', '[L3\g<MATCH1>Y', string, 1)
	elif action == "beta4Gal-T4":
		string = re.sub('\[(?P<MATCH1>(\[s6\])?)Y(?P<MATCH2>[36])', '[L4\g<MATCH1>Y\g<MATCH2>', string, 1)
	elif action == "ST3Gal-III/IV":
		string = re.sub('\[L(?P<MATCH1>([34])?)(?P<MATCH2>(\[f[34]\])?)(?P<MATCH3>(\[s6\])?)Y', '[S3L\g<MATCH1>\g<MATCH2>\g<MATCH3>Y', string, 1)
	elif action == "ST6Gal-I":
		string = re.sub('\[L4(?P<MATCH1>(\[f[34]\])?)(?P<MATCH2>(\[s6\])?)Y', '[S6L4\g<MATCH1>\g<MATCH2>Y', string, 1)
	elif action == "C2Gn-T":
		string = re.sub('^\[L3\]V(?P<MATCH1>[ST])', '[Y6][L3]V\g<MATCH1>', string, 1)
	elif action == "beta3Gn-T2/3/4/5/7":
		if advanced :
			string = re.sub('\[(?P<MATCH1>(\[s[6]\])?)L4(?P<MATCH2>(\[f[34]\])?)(?P<MATCH3>(\[s6\])?)Y', '[Y3\g<MATCH1>L4\g<MATCH2>\g<MATCH3>Y', string, 1)
		else :
			string = re.sub('\[(?P<MATCH1>(\[s[6]\])?)L4(?P<MATCH2>(\[s6\])?)Y', '[Y3\g<MATCH1>L4\g<MATCH2>Y', string, 1)
	elif action == "beta3Gn-T6":
		string = re.sub('^V(?P<MATCH1>[ST])', '[Y3]V\g<MATCH1>', string, 1)
	elif action == "C2/4Gn-T":
		string = re.sub('^\[Y3\]V(?P<MATCH1>[ST])', '[Y6][Y3]V\g<MATCH1>', string, 1)
	elif action == "alpha3Fuc-T":
		string = re.sub('\[L4(?P<MATCH1>(\[s6\])?)Y(?P<MATCH2>.)', '[L4[f3]\g<MATCH1>Y\g<MATCH2>', string, 1)
	elif action == "alpha4Fuc-T":
		string = re.sub('\[L3(?P<MATCH1>(\[s6\])?)Y', '[L3[f4]\g<MATCH1>Y', string, 1)
	elif action == "alpha2Fuc-Ts":
		string = re.sub('\[L(?P<MATCH1>[34])(?P<MATCH2>(\[f[34]\])?)(?P<MATCH3>(\[s6\])?)(?P<MATCH4>(\]V|Y))', '[[f2]L\g<MATCH1>\g<MATCH2>\g<MATCH3>\g<MATCH4>', string, 1)
	elif action == "Gcnt2":
		string = re.sub('\[(?P<MATCH1>(\[s[36]\])?)L4(?P<MATCH2>(\[s6\])?)Y3L(?P<MATCH3>(4|3\]V))', '[[Y6][\g<MATCH1>L4\g<MATCH2>Y3]L\g<MATCH3>', string, 1)
	elif action == "CHST4/6":
		string = re.sub('\[Y', '[[s6]Y', string, 1)
	elif action == "GALST2/3":
		string = re.sub('\[L3', '[[s3]L3', string, 1)
	elif action == "GALST4":
		string = re.sub('\[L4', '[[s3]L4', string, 1)
	elif action == "SAT-2":
		string = re.sub('\[S3(?P<MATCH1>(\[s[36]\])?)', '[S8S3\g<MATCH1>L', string, 1)
	elif action == "alpha3Gal-T":
		string = re.sub('\[\[f2\](?P<MATCH1>(\[s[36]\])?)L', '[La3[f2]\g<MATCH1>L', string, 1)
	elif action == "alpha3GalNAc-T":
		string = re.sub('\[\[f2\](?P<MATCH1>(\[s[36]\])?)L', '[Va3[f2]\g<MATCH1>L', string, 1)
	elif action == "beta4GalNAc-T":
		string = re.sub('\[S3(?P<MATCH1>(\[s[36]\])?)L', '[S3[Vb4]\g<MATCH1>L', string, 1);
	return string;

def chunk1( substrate, enzyme ):
	# for substrates with a single pattern-match
	global newg
	retstr = ""
	product = react(substrate,enzyme);
	if product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk1a( substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) with a single pattern-match
	global newg, newp, halt, limit
	retstr = ""
	product = react(substrate,enzyme);
	c = product.count('Y')
	if limit and limit > c :
		halt = 1
	if halt < 1 and product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		n[product] = len(pool)
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk2( regex, substrate, enzyme ):
	# for substrates having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l > 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				if nextproduct != '' and nextproduct != nextsubstrate :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1(substrate,enzyme)
	return

def chunk2a( regex, substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l > 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				c = product.count('Y')
				if limit and limit > c :
					halt = 1
				if halt < 1 and nextproduct != '' and nextsubstrate != nextproduct :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1a(substrate,enzyme)
	return

# start of main loop

while newg > 0 and iteration < maxiter :
	iteration += 1
	newg = 0

	news = newp	# the new products of one iteration become the substrates of the next
	newp = {}

	# cycle through the list of new substrates
	for substrate in news:

		# POLYPEPTIDE INITIATION
		enzyme = "ppGalNAc-Ts"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# SIALYL-Ts
		enzyme = "ST6GalNAc-I"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# T (CORE 1)
		enzyme = "C1Gal-T1"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# ST (CORE 1)
		enzyme = "ST3Gal-I"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# DIST (CORE 1)
		enzyme = "ST6GalNAc-III/IV"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# CORE 1 EXTENSION (CORE 1)
		enzyme = "beta3Gn-T3"
		if p[enzyme] > 0:
			chunk1a(substrate,enzyme);

		# ST6GlcNAc-I CORE 1 EXTENSION (CORE 1)
		enzyme = "ST6GlcNAc-I"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme);

		# beta3Gal-T5 
		enzyme = "beta3Gal-T5"
		if p[enzyme] > 0:
			if re.search('\[\[s6\]Y', substrate) :
				chunk2('(\[\[s6\]Y)',substrate,enzyme)
			if re.search('\[Y', substrate) :
				chunk2('(\[Y)',substrate,enzyme)

		# beta4Gal-T4 CORE 1 EXTENSION (CORE 1)
		enzyme = "beta4Gal-T4"
		if p[enzyme] > 0:
			if re.search('\[\[s6\]Y[36]', substrate) :
				chunk2('(\[\[s6\]Y[36])',substrate,enzyme)
			if re.search('\[Y[36]', substrate) :
				chunk2('(\[Y[36])',substrate,enzyme)

		# SIALYL-T3
		enzyme = "ST3Gal-III/IV"
		if p[enzyme] > 0:
			if re.search('\[L[34]\[f[34]\]\[s6\]Y', substrate) :
				chunk2('(\[L[34]\[f[34]\]\[s6\]Y)',substrate,enzyme)
			if re.search('\[L[34]\[f[34]\]Y', substrate) :
				chunk2('(\[L[34]\[f[34]\]Y)',substrate,enzyme)
			if re.search('\[L[34]\[s6\]Y', substrate) :
				chunk2('(\[L[34]\[s6\]Y)',substrate,enzyme)
			if re.search('\[L[34]Y', substrate) :
				chunk2('(\[L[34]Y)',substrate,enzyme)

		# SIALYL-T6 (TYPE-2 CHAIN)
		enzyme = "ST6Gal-I"
		if p[enzyme] > 0:
			if re.search('\[L4\[f[34]\]\[s6\]Y', substrate) :
				chunk2('(\[L4\[f[34]\]\[s6\]Y)',substrate,enzyme)
			if re.search('\[L4\[f[34]\]Y', substrate) :
				chunk2('(\[L4\[f[34]\]Y)',substrate,enzyme)
			if re.search('\[L4\[s6\]Y', substrate) :
				chunk2('(\[L4\[s6\]Y)',substrate,enzyme)
			if re.search('\[L4Y', substrate) :
				chunk2('(\[L4Y)',substrate,enzyme)

		# C2Gn-T (CORE 2)
		enzyme = "C2Gn-T"
		if p[enzyme] > 0:
			chunk1a(substrate,enzyme)


		# ELONGATION (CORE 2/4)
		enzyme = "beta3Gn-T2/3/4/5/7"
		if p[enzyme] > 0:
			# use 'and not...' to add further boolean restrictions on enzyme action
			if re.search('\[\[s[6]\]L4\[s6\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
				chunk2a('(\[\[s[6]\]L4\[s6\]Y)',substrate,enzyme)
			if re.search('\[L4\[s6\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
				chunk2a('(\[L4\[s6\]Y)',substrate,enzyme)
			if re.search('\[\[s[6]\]L4Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
				chunk2a('(\[\[s[6]\]L4Y)',substrate,enzyme)
			if re.search('\[L4Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
				chunk2a('(\[L4Y)',substrate,enzyme)
			if advanced :
				if re.search('\[\[s[6]\]L4\[f3\]\[s6\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
					chunk2a('(\[\[s[6]\]L4\[f3\]\[s6\]Y)',substrate,enzyme)
				if re.search('\[L4\[f3\]\[s6\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
					chunk2a('(\[L4\[f3\]\[s6\]Y)',substrate,enzyme)
				if re.search('\[\[s[6]\]L4\[f3\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
					chunk2a('(\[\[s[6]\]L4\[f3\]Y)',substrate,enzyme)
				if re.search('\[L4\[f3\]Y', substrate) and not re.search('Y6\]\[.+Y3\]L4', substrate):
					chunk2a('(\[L4\[f3\]Y)',substrate,enzyme)




		# beta3Gn-T6 (CORE 3)
		enzyme = "beta3Gn-T6"
		if p[enzyme] > 0:
			chunk1a(substrate,enzyme)

		# C2/4Gn-T (CORE 4)
		enzyme = "C2/4Gn-T"
		if p[enzyme] > 0:
			chunk1a(substrate,enzyme)

		# alpha3-FUCOSYLATION (Le[a/x])
		enzyme = "alpha3Fuc-T"
		if p[enzyme] > 0:
			if re.search('\[L4\[s6\]Y.', substrate) :
				chunk2('(\[L4\[s6\]Y.)',substrate,enzyme)
			if re.search('\[L4Y.', substrate) :
				chunk2('(\[L4Y.)',substrate,enzyme)

		# alpha4-FUCOSYLATION (Le[a/x])
		enzyme = "alpha4Fuc-T"
		if p[enzyme] > 0:
			if re.search('\[L3\[s6\]Y', substrate) :
				chunk2('(\[L3\[s6\]Y)',substrate,enzyme)
			if re.search('\[L3Y', substrate) :
				chunk2('(\[L3Y)',substrate,enzyme)

		# alpha2-FUCOSYLATION (Le[b/y])
		enzyme = "alpha2Fuc-Ts"
		if p[enzyme] > 0:
			if re.search('\[L[34]\[f[34]\]\[s6\]Y', substrate) :
				chunk2('(\[L[34]\[f[34]\]\[s6\]Y)',substrate,enzyme)
			if re.search('\[L[34]\[f[34]\]Y', substrate) :
				chunk2('(\[L[34]\[f[34]\]Y)',substrate,enzyme)
			if re.search('\[L[34]\[s6\]Y', substrate) :
				chunk2('(\[L[34]\[s6\]Y)',substrate,enzyme)
			if re.search('\[L[34]Y', substrate) :
				chunk2('(\[L[34]Y)',substrate,enzyme)
			# core 1/2 Gal support
			if re.search('\[L3\]V', substrate) :
				chunk1(substrate,enzyme)

		# type 2 double branching
		enzyme = "Gcnt2"
		if p[enzyme] > 0:
			if re.search('\[(\[s[36]\])?L4(\[s6\])?Y3L4', substrate) :
				if re.search('\[\[s[36]\]L4\[s6\]Y3L', substrate) :
					chunk2a('(\[\[s[36]\]L4\[s6\]Y3L)',substrate,enzyme)
				if re.search('\[L4\[s6\]Y3L4', substrate) :
					chunk2a('(\[L4\[s6\]Y3L4)',substrate,enzyme)
				if re.search('\[\[s[36]\]L4Y3L4', substrate) :
					chunk2a('(\[\[s[36]\]L4Y3L4)',substrate,enzyme)
				if re.search('\[L4Y3L4', substrate) :
					chunk2a('(\[L4Y3L4)',substrate,enzyme)
			if re.search('\[(\[s[36]\])?L4(\[s6\])?Y3L3\]V', substrate) :
				chunk1a(substrate,enzyme)

		# terminal GlcNAc sulfation
		enzyme = "CHST4/6"
		if p[enzyme] > 0:
			if re.search('\[Y', substrate) :
				chunk2('(\[Y)',substrate,enzyme)

		# terminal Gal 3-O-sulfation type 1
		enzyme = "GALST2/3"
		if p[enzyme] > 0:
			if re.search('\[L3', substrate) :
				chunk2('(\[L3)',substrate,enzyme)

		# terminal Gal 3-O-sulfation type 2
		enzyme = "GALST4"
		if p[enzyme] > 0:
			if re.search('\[L4', substrate) :
				chunk2('(\[L4)',substrate,enzyme)

		# alpha-2,8-sialyltransferase
		enzyme = "SAT-2"
		if p[enzyme] > 0:
			if re.search('\[S3\[s[36]\]L', substrate) :
				chunk2('(\[S3\[s[36]\]L)',substrate,enzyme)
			if re.search('\[S3L', substrate) :
				chunk2('(\[S3L)',substrate,enzyme)

		# alpha-1,3-galactosyltransferase
		enzyme = "alpha3Gal-T"
		if p[enzyme] > 0:
			if re.search('\[\[f2\]\[s[36]\]L', substrate) :
				chunk2('(\[\[f2\]\[s[36]\]L)',substrate,enzyme)
			if re.search('\[\[f2\]L', substrate) :
				chunk2('(\[\[f2\]L)',substrate,enzyme)

		# alpha-1,3-N-acetylgalactosaminyltransferase
		enzyme = "alpha3GalNAc-T"
		if p[enzyme] > 0:
			if re.search('\[\[f2\]\[s[36]\]L', substrate) :
				chunk2('(\[\[f2\]\[s[36]\]L)',substrate,enzyme)
			if re.search('\[\[f2\]L', substrate) :
				chunk2('(\[\[f2\]L)',substrate,enzyme)

		# beta-1,4-N-acetylgalactosaminyltransferase
		enzyme = "beta4GalNAc-T"
		if p[enzyme] > 0:
			if re.search('\[Sa?3\[s[36]\]Lb?4', substrate) :
				chunk2('(\[Sa?3\[s[36]\]Lb?4)',substrate,enzyme)
			if re.search('\[Sa?3Lb?4', substrate) :
				chunk2('(\[Sa?3Lb?4)',substrate,enzyme)


	if newg == 0 :
		print ("Nothing to be done!")
	else :
		print ("New structures: " + str(newg))

# print a list of all generated species

print ("\nO-glycan pool:")

for key in sorted(pool):
	print (key)

